﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Analysis.FitnessLandscape.DataTables;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Optimization;
using System.Collections.Generic;

namespace HeuristicLab.Analysis.FitnessLandscape.Analysis {

  [Item("Neutrality Analzyer", "Analyzes neutral parts of a trajectory")]
  [StorableClass]
  public class NeutralityAnalyzer : AlgorithmOperator, IAnalyzer {
    public bool EnabledByDefault {
      get { return false; }
    }

    #region Parameters
    public LookupParameter<DoubleValue> QualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<DoubleValue> CurrentNeutralDistanceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["CurrentNeutralDistance"]; }
    }
    public LookupParameter<DoubleValue> CurrentFractionOfNeutralNeighborsParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["CurrentFractionOfNeutralNeighbors"]; }
    }
    public LookupParameter<DataTable> NeutralWalkDistancesTableParameter {
      get { return (LookupParameter<DataTable>)Parameters["NeutralWalkDistancesTable"]; }
    }
    public LookupParameter<DoubleValue> AverageNeutralWalkLengthParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["AverageNeutralWalkLength"]; }
    }
    public LookupParameter<DoubleValue> AverageNeutralWalkDistanceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["AverageNeutralWalkDistance"]; }
    }
    public LookupParameter<DoubleValue> AverageNeutralNeighborFractionParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["AverageNeutralNeighborFraction"]; }
    }
    public LookupParameter<DoubleValue> AverageInnerNeutralNeighborFractionParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["AverageInnerNeutralNeighborFraction"]; }
    }
    public LookupParameter<DoubleValue> NeutralWalkLengthVarianceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["NeutralWalkLengthVariance"]; }
    }
    public LookupParameter<DoubleValue> NeutralWalkDistanceVarianceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["NeutralWalkDistanceVariance"]; }
    }        
    public LookupParameter<DoubleValue> NeutralNeighborFractionVarianceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["NeutralNeighborFractionVariance"]; }
    }
    public LookupParameter<DoubleValue> InnerNeutralNeighborFractionVarianceParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["InnerNeutralNeighborFractionVariance"]; }
    }
    public LookupParameter<VariableCollection> ResultsParameter {
      get { return (LookupParameter<VariableCollection>)Parameters["Results"]; }
    }
    #endregion

    [StorableConstructor]
    protected NeutralityAnalyzer(bool deserializing) : base(deserializing) { }
    protected NeutralityAnalyzer(NeutralityAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public NeutralityAnalyzer() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The current quality"));
      Parameters.Add(new LookupParameter<DoubleValue>("CurrentNeutralDistance", "The distance of the current solution to the starting point of the neutral (portion of the) walk."));
      Parameters.Add(new LookupParameter<DoubleValue>("CurrentFractionOfNeutralNeighbors", "The current fraction of neighbors with the same fitness (within epsilon)"));
      Parameters.Add(new LookupParameter<DataTable>("NeutralWalkDistancesTable", "The historical values of the current neutral distance."));
      Parameters.Add(new LookupParameter<DoubleValue>("AverageNeutralWalkLength", "The average length of a neutral walk"));
      Parameters.Add(new LookupParameter<DoubleValue>("AverageNeutralWalkDistance", "The average distance of the neutral walk to the starting point."));
      Parameters.Add(new LookupParameter<DoubleValue>("AverageNeutralNeighborFraction", "The average fraction of neutral neighbors overall."));
      Parameters.Add(new LookupParameter<DoubleValue>("AverageInnerNeutralNeighborFraction", "The average fraction of neutral neighbors within a neutral area."));
      Parameters.Add(new LookupParameter<DoubleValue>("NeutralWalkLengthVariance", "The variance of the neutral walks lengths"));
      Parameters.Add(new LookupParameter<DoubleValue>("NeutralWalkDistanceVariance", "The varaince of the neutral walks distances"));
      Parameters.Add(new LookupParameter<DoubleValue>("NeutralNeighborFractionVariance", "The average fraction of neutral neighbors overall."));
      Parameters.Add(new LookupParameter<DoubleValue>("InnerNeutralNeighborFractionVariance", "The average fraction of neutral neighbors within a neutral area."));

      var resultsCollector = new ResultsCollector();
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(CurrentNeutralDistanceParameter.ActualName));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(CurrentFractionOfNeutralNeighborsParameter.ActualName));
      resultsCollector.CollectedValues.Add(new LookupParameter<DataTable>(NeutralWalkDistancesTableParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(AverageNeutralWalkLengthParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(AverageNeutralWalkDistanceParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(AverageNeutralNeighborFractionParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(AverageInnerNeutralNeighborFractionParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(NeutralWalkLengthVarianceParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(NeutralWalkDistanceVarianceParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(NeutralNeighborFractionVarianceParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(InnerNeutralNeighborFractionVarianceParameter.Name));

      OperatorGraph.InitialOperator = resultsCollector;
      resultsCollector.Successor = null;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new NeutralityAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      DataTable distanceTable = GetOrCreateDistanceTable();
      var distanceValues = distanceTable.Rows["Neutral Walk Distances"].Values;
      var fractionValues = distanceTable.Rows["Neutral Neighbor Fractions"].Values;
      distanceValues.Add(CurrentNeutralDistanceParameter.ActualValue == null ? 0 : CurrentNeutralDistanceParameter.ActualValue.Value);
      fractionValues.Add(CurrentFractionOfNeutralNeighborsParameter.ActualValue == null ? 0 : CurrentFractionOfNeutralNeighborsParameter.ActualValue.Value);
      var lengths = new List<double>();
      var distances = new List<double>();
      var innerFractions = new List<double>();
      double lastValue = -1;
      int lastStart = 0;
      for (int i = 0; i<distanceValues.Count; i++) {
        if (distanceValues[i] < lastValue) { // restart
          lengths.Add(i - lastStart - 1);
          distances.Add(lastValue);
          lastStart = i;
        } else { // inside neutral area
          innerFractions.Add(fractionValues[i]);
        }
        lastValue = distanceValues[i];
      }
      if (lastStart != distanceValues.Count-1) {
        lengths.Add(distanceValues.Count - lastStart - 1);
        distances.Add(lastValue);
      }
      if (distances.Count > 0 && lengths.Count > 0) {
        AverageNeutralWalkDistanceParameter.ActualValue = new DoubleValue(distances.Average());
        NeutralWalkDistanceVarianceParameter.ActualValue = new DoubleValue(distances.Variance());
        AverageNeutralWalkLengthParameter.ActualValue = new DoubleValue(lengths.Average());
        NeutralWalkLengthVarianceParameter.ActualValue = new DoubleValue(lengths.Variance());
        AverageNeutralNeighborFractionParameter.ActualValue = new DoubleValue(fractionValues.Average());
        NeutralNeighborFractionVarianceParameter.ActualValue = new DoubleValue(fractionValues.Variance());
        AverageInnerNeutralNeighborFractionParameter.ActualValue = new DoubleValue(innerFractions.Average());
        InnerNeutralNeighborFractionVarianceParameter.ActualValue = new DoubleValue(innerFractions.Variance());
      }
      return base.Apply();
    }

    private DataTable GetOrCreateDistanceTable() {
      DataTable distancesTable = NeutralWalkDistancesTableParameter.ActualValue;
      if (distancesTable == null) {
        distancesTable = new DataTable("Neutral Walk Distances");
        NeutralWalkDistancesTableParameter.ActualValue = distancesTable;
        distancesTable.Rows.Add(new DataRow("Neutral Walk Distances"));
        distancesTable.Rows.Add(new DataRow("Neutral Neighbor Fractions") {VisualProperties = {SecondYAxis = true}});
      }
      return distancesTable;
    }
  }
}
